from abc import ABCMeta

from .utils.conformance import verify_conformance, verify_not_overridden
from .utils.docs import update_docs
from .utils.inspection import (
    set_explicit_override,
    set_forced_override,
    set_quirk_docs_method,
    set_quirk_docs_mro,
    should_skip,
)


class InterfaceMeta(ABCMeta):
    """
    A metaclass that helps subclasses of a class to conform to its API.

    It also makes sure that documentation that might be useful to a user
    is inherited appropriately, and provides a hook for class to handle
    subclass operations.
    """

    INTERFACE_EXPLICIT_OVERRIDES = True
    INTERFACE_RAISE_ON_VIOLATION = False
    INTERFACE_SKIPPED_NAMES = set()

    def __init__(cls, name, bases, dct):
        ABCMeta.__init__(cls, name, bases, dct)

        # Register interface class for subclasses
        if not hasattr(cls, "__interface__"):
            cls.__interface__ = cls

        # Read configuration
        explicit_overrides = cls.__get_config(
            bases, dct, "INTERFACE_EXPLICIT_OVERRIDES"
        )
        raise_on_violation = cls.__get_config(
            bases, dct, "INTERFACE_RAISE_ON_VIOLATION"
        )
        skipped_names = cls.__get_config(bases, dct, "INTERFACE_SKIPPED_NAMES")

        # Iterate over names in `dct` and check for conformance to interface
        for key, value in dct.items():

            # Skip any key corresponding to Python magic methods
            if key.startswith("__") and key.endswith("__"):
                continue

            # Skip any key in skipped_names
            if key in skipped_names or should_skip(value):  # pragma: no cover
                continue

            # Identify the first instance of this key in the MRO, if it exists, and check conformance
            is_override = False
            for base in cls.__mro__[1:]:
                if base is object:
                    continue
                if key in base.__dict__:
                    is_override = True
                    cls.__verify_conformance(
                        key,
                        name,
                        value,
                        base.__name__,
                        base.__dict__[key],
                        explicit_overrides=explicit_overrides,
                        raise_on_violation=raise_on_violation,
                    )
                    break
                if key in getattr(
                    base, "__annotations__", {}
                ):  # Declared but as yet unspecified attributes
                    is_override = True
                    cls.__verify_conformance(
                        key,
                        name,
                        value,
                        name,
                        None,
                        explicit_overrides=explicit_overrides,
                        raise_on_violation=raise_on_violation,
                    )
                    break

            if not is_override:
                verify_not_overridden(
                    key, name, value, raise_on_violation=raise_on_violation
                )

        # Update documentation
        cls.__update_docs(cls, name, bases, dct)

        # Call subclass registration hook
        cls.__register_implementation__()

    def __register_implementation__(cls):
        pass

    @classmethod
    def __get_config(mcls, bases, dct, key):
        default = getattr(mcls, key, None)
        if bases:
            default = getattr(bases[0], key, default)
        return dct.get(key, default)

    @classmethod
    def __verify_conformance(
        mcls,
        key,
        name,
        value,
        base_name,
        base_value,
        explicit_overrides=True,
        raise_on_violation=False,
    ):
        return verify_conformance(
            key,
            name,
            value,
            base_name,
            base_value,
            explicit_overrides=explicit_overrides,
            raise_on_violation=raise_on_violation,
        )

    @classmethod
    def __update_docs(mcls, cls, name, bases, dct):
        skipped_names = mcls.__get_config(bases, dct, "INTERFACE_SKIPPED_NAMES")
        return update_docs(cls, name, bases, dct, skipped_names=skipped_names)

    @classmethod
    def inherit_docs(mcls, method=None, mro=True):
        """
        Indicate to `InterfaceMeta` how the wrapped method should be documented.

        Methods need not normally be decorated with this decorator, except in the
        following cases:
        - documentation for quirks should be lifted not from overrides to a
        method, but from some other method (e.g. because subclasses or
        implementations of the interface should override behaviour in a "private"
        method rather than the top-level public method).
        - the method has been nominated by the interface configuration to be
        skipped, in which case decorating with this method will enable
        documentation generation as if it were not.

        Use this decorator as `@InterfaceMeta.inherit_docs([method=...], [mro=...])`
        or `@<metaclass instance>.inherit_docs([method=...], [mro=...])`.

        Args:
            method (str, None): A method from which documentation for implementation
                specific quirks should be extracted. [Useful when implementations
                of an interface are supposed to change underlying methods rather
                than the public method itself].
            mro (bool): Whether to include documentation from all levels of the
                MRO, starting from the most primitive class that implementated it.
                All higher levels will be considered as "quirks" to the interface's
                definition.

        Returns:
            function: A function wrapper that attaches attributes `_quirks_method` and
            `_quirks_mro` to the method, for interpretation by `InterfaceMeta`.
        """

        def doc_wrapper(f):
            set_quirk_docs_method(f, method)
            set_quirk_docs_mro(f, mro)
            return f

        return doc_wrapper

    @classmethod
    def override(mcls, func=None, force=False):
        """
        Indicate to `InterfaceMeta` that this method has intentionally overridden an interface method.

        This decorator also allows one to indicate that the method should be
        overridden without warnings even when it does not conform to the API.

        Use this decorator as `@InterfaceMeta.override`, `@InterMeta.override(force=True)`,
        `@<metaclass instance>.override`, or `@<metaclass instance>.override(force=True)`.

        A recommended convention is to use this decorator as the outermost decorator.

        Args:
            f (function, None): The function, if method is decorated by the decorator
                without arguments (e.g. @override), else None.
            force (bool): Whether to force override of method even if the API does
                note match. Note that in this case, documentation is not inherited
                from the MRO.

        Returns:
            function: The wrapped function of function wrapper depending on which
                arguments are present.
        """

        def override(f):
            set_explicit_override(f)
            set_forced_override(f, force)
            return f

        if func is not None:
            return override(func)
        return override



import inspect
import logging
from abc import abstractproperty
from inspect import Parameter

from .inspection import (
    get_functional_signature,
    has_explicit_override,
    has_forced_override,
    is_functional_member,
    is_method,
    should_skip,
)
from .reporting import report_violation


def verify_conformance(
    name,
    clsname,
    member,
    ref_clsname,
    ref_member,
    explicit_overrides=True,
    raise_on_violation=False,
):
    """
    Verify that a member conforms to a nominated interface.

    Args:
        name (str): The name of the member method being checked.
        clsname (str): The name of the class parent of the checked member.
        member (object): The class member to check for conformance against
            ref_member.
        ref_clsname (str): The name of the reference class to be treated as an
            interface.
        ref_member (object): The referece member to be treated as an interface
            definition.
        explicit_overrides (bool): Whether to require explicit overrides.
            (default: True)
        raise_on_violation (bool): Whether any non-conformance should cause an
            exception to be raised. (default: False)
    """
    if hasattr(
        ref_member, "__objclass__"
    ):  # pragma: no cover; Method is attached to metaclass, so should not be checked.
        return

    if has_forced_override(member) or should_skip(ref_member):
        return

    # Check that type of member has not changed.
    if type(member) is not type(ref_member):
        if type(member) in (abstractproperty, property) and not is_method(ref_member):
            # This should be okay, provided the property is properly crafted.
            pass
        elif is_functional_member(member):
            # This means we are replacing a fixed attribute with a method,
            # or between different types of functional members
            report_violation(
                "`{}.{}` changes the type of `{}.{}` (`{}` instead of `{}`) without using `@override(force=True)` decorator.".format(
                    clsname, name, ref_clsname, name, type(member), type(ref_member)
                ),
                raise_on_violation,
            )
        else:  # Most other type changes should be fine
            pass

    # Check that overrides are present
    if (
        is_functional_member(member)
        or inspect.isdatadescriptor(member)
        or inspect.ismethoddescriptor(member)
    ):
        if explicit_overrides and not has_explicit_override(member):
            report_violation(
                "`{}.{}` overrides interface `{}.{}` without using the `@override` decorator.".format(
                    clsname, name, ref_clsname, name
                ),
                raise_on_violation,
            )

    if is_functional_member(member) and is_functional_member(ref_member):
        verify_signature(
            name,
            clsname,
            member,
            ref_clsname,
            ref_member,
            raise_on_violation=raise_on_violation,
        )


def verify_signature(
    name, clsname, member, ref_clsname, ref_member, raise_on_violation=False
):
    """
    Verify that the signature of a member is compatible with some reference member.

    Args:
        name (str): The name of the member method being checked.
        clsname (str): The name of the class parent of the checked member.
        member (object): The class member to check for conformance against
            ref_member.
        ref_clsname (str): The name of the reference class to be treated as an
            interface.
        ref_member (object): The referece member to be treated as an interface
            definition.
        raise_on_violation (bool): Whether any non-conformance should cause an
            exception to be raised. (default: False)
    """
    sig = get_functional_signature(member)
    ref_sig = get_functional_signature(ref_member)

    if not check_signatures_compatible(sig, ref_sig):
        message = "Signature `{}.{}{}` does not conform to interface `{}.{}{}`.".format(
            clsname, name, sig, ref_clsname, name, ref_sig
        )
        if raise_on_violation:
            raise RuntimeError(message)
        else:
            logging.warning(message)


def check_signatures_compatible(sig, ref_sig):
    """
    Check whether two signatures are compatible.

    Args:
        sig (Signature): A signature of a member to check for compatibility with
            `ref_sig`.
        ref_sig (Signature): The reference signature.

    Returns:
        bool: `True` if the signatures are compatible, and `False` otherwise.
    """
    params = iter(sig.parameters.values())
    base_params = iter(ref_sig.parameters.values())

    try:
        for bp in base_params:
            cp = next(params)

            while (
                bp.kind is Parameter.VAR_POSITIONAL
                and cp.kind is Parameter.POSITIONAL_OR_KEYWORD
            ):
                cp = next(params)

            while (
                bp.kind is Parameter.VAR_KEYWORD
                and cp.kind is not Parameter.VAR_KEYWORD
            ):
                cp = next(params)

            if not (
                cp.name == bp.name and bp.kind == cp.kind and bp.default == cp.default
            ):
                raise ValueError(bp, cp)

    except (StopIteration, ValueError):
        return False

    for param in params:
        if (
            param.kind is Parameter.POSITIONAL_ONLY
            or param.kind is Parameter.POSITIONAL_OR_KEYWORD
            and param.default == inspect._empty
        ):
            return False

    return True


def verify_not_overridden(name, clsname, member, raise_on_violation=False):
    """
    Verify that a nominated member is *not* an override.

    Args:
        name (str): The name of the member method being checked.
        clsname (str): The name of the class parent of the checked member.
        member (object): The class member to check for conformance against
            ref_member.
        raise_on_violation (bool): Whether any non-conformance should cause an
            exception to be raised. (default: False)
    """
    if has_explicit_override(member):
        report_violation(
            "`{clsname}.{name}` claims to override interface method, but no such method exists.".format(
                clsname=clsname, name=name
            ),
            raise_on_violation=raise_on_violation,
        )



import logging

from .errors import InterfaceConformanceError


def report_violation(message, raise_on_violation):
    """
    Report a violation in conformance to the user.

    Args:
        message (str): The message to pass on to the user.
        raise_on_violation (bool): Whether any non-conformance should cause an
            exception to be raised. (default: False)
    """
    if raise_on_violation:
        raise InterfaceConformanceError(message)
    else:
        logging.warning(message)



import inspect
import textwrap
from collections import OrderedDict

from .inspection import (
    get_class_attr_docs,
    get_functional_docs,
    get_functional_wrapper,
    get_quirk_docs_method,
    get_quirk_docs_mro,
    has_class_attr_docs,
    has_forced_override,
    has_quirk_docs_mro,
    has_updatable_docs,
    set_functional_docs,
)


def update_docs(cls, name, bases, dct, skipped_names=None):
    """
    Update the documentation on class members with information from parents.

    If a method is implementing an interface, its documentation should only need
    to include the quirks associated with the implementation, if any; and all
    other documentation should be inherited. This function performs all of the
    magic required to see that this is the case.

    Args:
        name (str): The name of the class being constructed.
        bases (list<class,type>): The bases of the class being constructed.
        dct (dict): The class dictionary being used to construct the class.
        skipped_names (list<str>): Names for which to skip the documentation
            rewriting.
    """

    mro = inspect.getmro(cls)
    mro = mro[: mro.index(cls.__interface__) + 1]
    skipped_names = skipped_names or set()

    # Handle module-level documentation
    module_docs = [cls.__doc__]
    for klass in mro:
        if has_class_attr_docs(klass):
            module_docs.append(
                [
                    "Attributes:"
                    if klass is cls
                    else "Attributes inherited from {}:".format(klass.__name__),
                    inspect.cleandoc(get_class_attr_docs(klass)),
                ]
            )

    cls.__doc__ = doc_join(*module_docs)

    # Assemble class attribute names avoiding dunder methods
    members = {}
    for klass in reversed(cls.mro()):
        members.update(
            {
                name: member
                for name, member in klass.__dict__.items()
                if not name.startswith("__") and not name.endswith("__")
            }
        )

    # Handle function/method-level documentation
    for name, member in members.items():

        # Skip magic methods
        if name.startswith("__") and name.endswith("__"):
            continue

        # Check if there is anything to do
        if not has_updatable_docs(member):
            continue

        quirks_method = get_quirk_docs_method(member)
        quirks_mro = get_quirk_docs_mro(member)
        has_quirks_mro = has_quirk_docs_mro(member)

        if (
            inspect.isabstract(member)
            or has_forced_override(member)
            or name in skipped_names
            and not (has_quirks_mro or quirks_method)
            or name not in cls.__dict__
            and quirks_method is None
        ):
            continue

        # Extract documentation from this member and the quirks member
        method_docs = OrderedDict()
        last_docs = None
        for i, klass in enumerate(reversed(mro) if quirks_mro else mro[:1]):
            klass_member = klass.__dict__.get(name, None)
            if klass_member is not None:
                member_docs = get_functional_docs(klass_member)
                if (i == 0 or member_docs) and member_docs != last_docs:
                    last_docs = method_docs[klass.__name__] = member_docs
                if not get_quirk_docs_mro(klass_member):
                    break

        if quirks_method is not None and quirks_method in members:
            quirk_member = members.get(quirks_method)
            quirk_member_docs = get_functional_docs(quirk_member)
            if quirk_member_docs:
                if cls.__name__ in method_docs:
                    method_docs[cls.__name__] = (
                        inspect.cleandoc(method_docs[cls.__name__])
                        + "\n\n"
                        + inspect.cleandoc(quirk_member_docs)
                    )
                else:
                    method_docs[cls.__name__] = quirk_member_docs

        if method_docs:

            if name not in cls.__dict__:
                # Overide method object with new object so we don't modify
                # underlying method that may be shared by multiple classes.
                member = get_functional_wrapper(member)

            set_functional_docs(
                member,
                doc_join(
                    *[
                        docs if i == 0 else [source + " Quirks:", docs]
                        for i, (source, docs) in enumerate(method_docs.items())
                    ]
                ),
            )

            if name not in cls.__dict__:
                setattr(cls, name, member)


def doc_join(*docs):
    """
    Stitch multiple pieces of documentation into one docstring.

    Args:
        *docs (tuple<str, list<str>, tuple<str>>): A sequence of strings or
            length-2 sequences of strings to stitch together into a docstring.
            If a length-2 sequence is provided, then the first string is treated
            as a section header, and the rest of the string is indented
            beneath it.

    Returns:
        str: The stitched together docstring.
    """
    out = []
    for doc in docs:
        if doc in (None, ""):
            continue
        elif isinstance(doc, str):
            out.append(textwrap.dedent(doc).strip("\n"))
        elif isinstance(doc, (list, tuple)):
            if len(doc) < 2:
                continue
            d = doc_join(*doc[1:])
            if d:
                if not out:
                    out.append("\n")
                out.append(
                    "{header}\n{body}".format(
                        header=doc[0].strip(),
                        body="    "
                        + d.replace(
                            "\n", "\n    "
                        ),  # textwrap.indent not available in python2
                    )
                )
        else:
            raise ValueError("Unrecognised doc format: {}".format(type(doc)))
    return "\n\n".join(out) or None



import warnings

from .interface import InterfaceMeta
from .utils.inspection import set_skip


def inherit_docs(method=None, mro=True):
    """
    Indicate to `InterfaceMeta` how the wrapped method should be documented.

    Methods need not normally be decorated with this decorator, except in the
    following cases:
    - documentation for quirks should be lifted not from overrides to a
      method, but from some other method (e.g. because subclasses or
      implementations of the interface should override behaviour in a "private"
      method rather than the top-level public method).
    - the method has been nominated by the interface configuration to be
      skipped, in which case decorating with this method will enable
      documentation generation as if it were not.

    Use this decorator as `@inherit_docs([method=...], [mro=...])`.

    Args:
        method (str, None): A method from which documentation for implementation
            specific quirks should be extracted. [Useful when implementations
            of an interface are supposed to change underlying methods rather
            than the public method itself].
        mro (bool): Whether to include documentation from all levels of the
            MRO, starting from the most primitive class that implementated it.
            All higher levels will be considered as "quirks" to the interface's
            definition.

    Returns:
        function: A function wrapper that attaches attributes `_quirks_method` and
        `_quirks_mro` to the method, for interpretation by `InterfaceMeta`.
    """
    return InterfaceMeta.inherit_docs(method=method, mro=mro)


def quirk_docs(method=None, mro=True):
    """
    DEPRECATED: Please use `inherit_docs` instead.
    """
    warnings.warn(
        "The `interface_meta.quirk_docs` decorator has been replaced by `implemented_by` and "
        "will be removed in version 2.0.",
        DeprecationWarning,
    )
    return inherit_docs(method=method, mro=mro)


def override(func=None, force=False, f=None):
    """
    Indicate to `InterfaceMeta` that this method has intentionally overridden an interface method.

    This decorator also allows one to indicate that the method should be
    overridden without warnings even when it does not conform to the API.

    Use this decorator as `@override` or `@override(force=True)`.

    A recommended convention is to use this decorator as the outermost decorator.

    Args:
        func (function, None): The function, if method is decorated by the decorator
            without arguments (e.g. @override), else None.
        force (bool): Whether to force override of method even if the API does
            note match. Note that in this case, documentation is not inherited
            from the MRO.
        f (function, None): Deprecated predecessor of `func`. Maintained for
            backward compatibility until v2.0.

    Returns:
        function: The wrapped function of function wrapper depending on which
            arguments are present.
    """
    if f is not None:
        warnings.warn(
            "The `f` argument to the `interface_meta.override` decorator has been renamed `func`. This "
            "backward compatibility shim will be removed in 2.0.",
            DeprecationWarning,
        )
    return InterfaceMeta.override(func=func or f, force=force)


def skip(func):
    """
    Indicate to `InterfaceMeta` that this method should be skipped.

    Args:
        func (function): The function/member to mark as skipped.

    Returns:
        function: The marked function/member (same instance as that passed in).
    """
    set_skip(func, skip=True)
    return func



import functools
import inspect
from abc import abstractproperty

try:
    from abc import abstractclassmethod, abstractstaticmethod
except ImportError:  # pragma: no cover
    abstractclassmethod = classmethod
    abstractstaticmethod = staticmethod

try:
    from inspect import signature
except ImportError:  # pragma: no cover
    from funcsigs import signature


# Abstract away differences between functions, methods and descriptors


def _get_member(member):
    if inspect.ismethod(member):
        return member.__func__
    if inspect.isdatadescriptor(member) and isinstance(
        member, (property, abstractproperty)
    ):
        return member.fget
    if inspect.ismethoddescriptor(member):
        if isinstance(member, (classmethod, abstractclassmethod)):
            return member.__get__(object, object).__func__
        if isinstance(member, (staticmethod, abstractstaticmethod)):
            return member.__get__(object, object)
    return member


def is_method(member):
    return inspect.isfunction(member) or inspect.ismethod(member)


def is_functional_member(member):
    """
    Check whether a class member from the __dict__ attribute is a method.

    This can be true in two ways:
        - It is literally a Python function
        - It is a method descriptor (wrapping a function)

    Args:
        member (object): An object in the class __dict__.

    Returns:
        bool: `True` if the member is a function (or acts like one).
    """
    return inspect.isfunction(member) or (
        inspect.ismethoddescriptor(member)
        and isinstance(member, (classmethod, staticmethod))
    )


def get_functional_signature(member):
    return signature(_get_member(member))


def functional_hasattr(member, attr):
    return hasattr(_get_member(member), attr)


def functional_getattr(member, attr, default=None):
    return getattr(_get_member(member), attr, default)


def functional_setattr(member, attr, value):
    setattr(_get_member(member), attr, value)


def functional_delattr(member, attr):
    delattr(_get_member(member), attr)


def get_functional_wrapper(member):
    """
    Return a new functional wrapper around the provided member.

    Since `interface_meta` rewrites documentation strings, if the affected
    method is from a parent class, then we need to create a new function wrapper
    so that mutating the documentation does not leak to higher classes. This
    function does just that.

    Args:
        member (functional): The functional object to be wrapped.

    Returns:
        functional: An object that is functionally equivalent to `member`,
            but which can have its own attributes.
    """
    class_method = isinstance(member, classmethod)
    property_method = isinstance(member, property)
    static_method = isinstance(member, staticmethod)

    function = _get_member(member)

    @functools.wraps(function)
    def wrapper(*args, **kwargs):  # pragma: no cover
        return function(*args, **kwargs)

    for attr in ["_quirks_method", "_quirks_mro", "__override__", "__override_force__"]:
        if hasattr(function, attr):
            setattr(wrapper, attr, getattr(function, attr))

    if class_method:
        return classmethod(wrapper)
    elif property_method:
        return property(fget=wrapper, fset=member.fset, fdel=member.fdel)
    elif static_method:
        return staticmethod(wrapper)
    return wrapper


# Override checking


def has_explicit_override(member):
    return functional_getattr(member, "__override__", False)


def set_explicit_override(member, override=True):
    return functional_setattr(member, "__override__", override)


def has_forced_override(member):
    return functional_getattr(member, "__override_force__", False)


def set_forced_override(member, force=True):
    return functional_setattr(member, "__override_force__", force)


# Skip interface conformance checks


def should_skip(member):
    return functional_getattr(member, "__interface_meta_skip__", False)


def set_skip(member, skip=True):
    return functional_setattr(member, "__interface_meta_skip__", skip)


# Documentation helpers


def has_updatable_docs(member):
    return (
        is_functional_member(member)
        or inspect.isdatadescriptor(member)
        and isinstance(member, (property, abstractproperty))
    )


def get_functional_docs(member, orig=True):
    if orig and functional_hasattr(member, "__doc_orig__"):
        return functional_getattr(member, "__doc_orig__")
    return functional_getattr(member, "__doc__")


def set_functional_docs(member, docs):
    if not functional_hasattr(member, "__doc_orig__"):
        functional_setattr(
            member, "__doc_orig__", functional_getattr(member, "__doc__")
        )
    functional_setattr(member, "__doc__", docs)


def has_class_attr_docs(cls):
    return hasattr(cls, "_{}__doc_attrs".format(cls.__name__))


def get_class_attr_docs(cls):
    return getattr(cls, "_{}__doc_attrs".format(cls.__name__))


# Quirk documentation helpers


def has_quirk_docs_method(member):
    return functional_hasattr(member, "_quirks_method")


def get_quirk_docs_method(member):
    return functional_getattr(member, "_quirks_method", None)


def set_quirk_docs_method(member, method):
    return functional_setattr(member, "_quirks_method", method)


def has_quirk_docs_mro(member):
    return functional_hasattr(member, "_quirks_mro")


def get_quirk_docs_mro(member):
    return functional_getattr(member, "_quirks_mro", True)


def set_quirk_docs_mro(member, mro):
    return functional_setattr(member, "_quirks_mro", mro)



from ._version import __author__, __author_email__, __version__  # noqa: F401

from .interface import InterfaceMeta
from .decorators import inherit_docs, override, quirk_docs, skip

__all__ = [
    # Interface
    "InterfaceMeta",
    # Standalone decorators
    "inherit_docs",
    "override",
    "skip",
    # Deprecated (will be removed in v2)
    "quirk_docs",
]
